#include <math.h>
#include <stdio.h>

#include <mpi.h>

#include "simple-jpeg/import_export_jpeg.h"
#include "mpi-utils.h"

int main (int argc, char * argv[]) {
  /* general */
  int i, j;
  int m, n;

  const char * infile;

  /* MPI specific */
  int mpi_size;
  int mpi_rank;

  MPI_Datatype mpi_image_row;

  int * scatter_counts;
  int * scatter_displs;

  /* local to each process */
  unsigned char * local_image_data;
  float local_sum  = 0.0;
  float local_mean = 0.0;
  float local_var  = 0.0;

  int local_m;
  int local_n;

  /* significant only at root */
  unsigned char * root_image_data = NULL;
  float root_std;

  int bytes_per_pixel;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);

  DEBUG_PID_ATTACH;

  if (argc < 2) {
    root_eprintf(mpi_rank,
		 "usage: %s image\n",
		 argv[0]);
    FINALIZE_AND_FAIL;
  }

  infile = argv[1];

  if (PROCESS_IS_ROOT(mpi_rank))
    import_JPEG_file(infile,
		     &root_image_data,
		     &m, &n,
		     &bytes_per_pixel);

  /* get everyone to agree on m and n
     before we create the datatype */
  MPI_Bcast(&m, 1, MPI_INT, ROOT_PROCESS, MPI_COMM_WORLD);
  MPI_Bcast(&n, 1, MPI_INT, ROOT_PROCESS, MPI_COMM_WORLD);

  /* create a row */
  MPI_Type_vector(1, n, n, MPI_UNSIGNED_CHAR, &mpi_image_row);
  MPI_Type_commit(&mpi_image_row);

  /* allocate work division matrices */
  scatter_counts = malloc(mpi_size * sizeof(int));
  scatter_displs = malloc(mpi_size * sizeof(int));
  
  if (scatter_counts == NULL ||
      scatter_displs == NULL) {
    fprintf(stderr,
	    "process %d: allocation failed.\n",
	    mpi_rank);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  /* work division */
  for (i = 0; i < mpi_size; i++) {
    scatter_counts[i] = BLOCK_SIZE(i, mpi_size, m);
    scatter_displs[i] = BLOCK_LOW (i, mpi_size, m);
  }

  local_m = scatter_counts[mpi_rank];
  local_n = n;

  /* allocate my share of the data */
  local_image_data =
    malloc(local_m * local_n * sizeof(unsigned char));

  if (local_image_data == NULL) {
    fprintf(stderr,
	    "process %d: allocation failed\n",
	    mpi_rank);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  /* get share */
  MPI_Scatterv(root_image_data,
	       scatter_counts, scatter_displs, mpi_image_row,
	       local_image_data, local_m, mpi_image_row,
	       ROOT_PROCESS, MPI_COMM_WORLD);

  /* calculate local sum */
  for (i = 0; i < local_m; i++)
    for (j = 0; j < local_n; j++)
      local_sum +=
	(float) local_image_data[i*local_n + j];

  /* everybody shares it */
  MPI_Allreduce(&local_sum, &local_mean,
		1, MPI_FLOAT, MPI_SUM,
		MPI_COMM_WORLD);
  local_mean /= m * n; /* average */

  /* calculate variance */
  for (i = 0; i < local_m; i++)
    for (j = 0; j < local_n; j++)
      local_var += (float)
	(local_image_data[i*local_n + j] - local_mean) *
	(local_image_data[i*local_n + j] - local_mean);

  /* root gets the sum of every variance */
  MPI_Reduce(&local_var, &root_std,
	     1, MPI_FLOAT, MPI_SUM,
	     ROOT_PROCESS, MPI_COMM_WORLD);

  /* root calculates standard deviation
     and prints everything to screen */
  if (PROCESS_IS_ROOT(mpi_rank)) {
    root_std = sqrtf(root_std/(n*m-1));
    printf("The average pixel value is: %f\n"
	   "The standard deviation is: %f\n",
	   local_mean,
	   root_std);
    fflush(stdout);
  }

  /* cleanup */
  free(local_image_data);
  free(scatter_displs);
  free(scatter_counts);

  MPI_Type_free(&mpi_image_row);

  free(root_image_data);

  FINALIZE_AND_SUCCEED;
}
