#include <stdio.h>
#include <mpi.h>

#include "dalloc.h"
#include "mpi-utils.h"

int main (int argc, char * argv[]) {
  /* general */
  int i, j;
  int P, Q;
  int m, n;

  int ** matrix;

  int synch = -1;

  /* mpi specific */

  MPI_Datatype mpi_col;
  MPI_Datatype mpi_row;

  int mpi_size;
  int mpi_rank;

  /* local data */
  int local_m;
  int local_n;
  int start_m;
  int start_n;
  int end_m;
  int end_n;

  int neighbor_l;
  int neighbor_r;
  int neighbor_d;
  int neighbor_u;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);

  DEBUG_PID_ATTACH;

  if (argc < 5) {
    root_eprintf(mpi_rank,
		 "usage: %s P Q m n\n",
		 argv[0]);
    FINALIZE_AND_FAIL;
  }

  P = atoi(argv[1]);
  Q = atoi(argv[2]);
  m = atoi(argv[3]);
  n = atoi(argv[4]);  

  /* sanity */
  if (P*Q != mpi_size ||
      m < 1 || n < 1 || P < 1 || Q < 1) {
    root_eprintf(mpi_rank,
		 "nonsensical values: P=%d, Q=%d, m=%d, n=%d\n",
		 P, Q, m, n);
    FINALIZE_AND_FAIL;
  }

  /* work division */
  local_m = BLOCK_SIZE(mpi_rank / Q, P, m-2) + 2;
  local_n = BLOCK_SIZE(mpi_rank % Q, Q, n-2) + 2;

  start_m = BLOCK_LOW(mpi_rank / Q, P, m-2);
  start_n = BLOCK_LOW(mpi_rank % Q, Q, n-2);

  end_m = local_m + start_m;
  end_n = local_n + start_n;

  /* which neighbors really exist? */
  neighbor_u = start_m    ? mpi_rank-Q : MPI_PROC_NULL;
  neighbor_d = end_m != m ? mpi_rank+Q : MPI_PROC_NULL;
  neighbor_l = start_n    ? mpi_rank-1 : MPI_PROC_NULL;
  neighbor_r = end_n != n ? mpi_rank+1 : MPI_PROC_NULL;

  /* data type setup, every neighbor
     agrees on rows and columns now */
  MPI_Type_vector(local_m, 1, local_n, MPI_INT, &mpi_col);
  MPI_Type_commit(&mpi_col);

  MPI_Type_vector(1, local_n, local_n, MPI_INT, &mpi_row);
  MPI_Type_commit(&mpi_row);

  /* allocate matrix */
  matrix = dalloc(local_m, local_n, sizeof(int));

  if (matrix == NULL) {
    fprintf(stderr,
	    "process %d: allocation failed\n",
	    mpi_rank);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  /* every element my local
     matrix is mpi_rank */
  for (i = 0; i < local_m; i++)
    for (j = 0; j < local_n; j++)
      matrix[i][j] = mpi_rank;

  /* send next to right-most column right
     receive column from left into left-most */
  MPI_Sendrecv(&matrix[0][local_n-2], 1, mpi_col, neighbor_r, 101,
  	       &matrix[0][0], 1, mpi_col, neighbor_l, 101,
  	       MPI_COMM_WORLD, MPI_STATUS_IGNORE);

  /* send next to bottom row down
     receive row from top into top */
  MPI_Sendrecv(matrix[local_m-2], 1, mpi_row, neighbor_d, 102,
  	       matrix[0], 1, mpi_row, neighbor_u, 102,
  	       MPI_COMM_WORLD, MPI_STATUS_IGNORE);

  /* synchronized print to screen */
  MPI_Recv(&synch, 1, MPI_INT,
  	   VALID_PROCESS(mpi_rank-1, mpi_size), 103,
  	   MPI_COMM_WORLD, MPI_STATUS_IGNORE);

  printf("%d:\n", mpi_rank);
  for (i = 0; i < local_m; i++) {
    for (j = 0; j < local_n; j++)
      printf("%2d, ", matrix[i][j]);
    puts("");
  }
  fflush(stdout);

  MPI_Send(&synch, 1, MPI_INT,
  	   VALID_PROCESS(mpi_rank+1, mpi_size), 103,
  	   MPI_COMM_WORLD);

  /* cleanup */
  dalloc_free(matrix);
  MPI_Type_free(&mpi_row);
  MPI_Type_free(&mpi_col);

  FINALIZE_AND_SUCCEED;
}
